from setuptools import setup
import os
import sys


def running_under_virtualenv():
    if hasattr(sys, "real_prefix"):
        return True
    elif sys.prefix != getattr(sys, "base_prefix", sys.prefix):
        return True
    if os.getenv("VIRTUAL_ENV", False):
        return True
    if "--user" in sys.argv:
        return True
    return False


VIRTUAL_ENV = running_under_virtualenv()


def get_dir(system_path=None, virtual_path=None):
    """
    Retrieve VIRTUAL_ENV friendly path
    :param system_path: Relative system path
    :param virtual_path: Overrides system_path for virtual_env only
    :return: VIRTUAL_ENV friendly path
    """
    if virtual_path is None:
        virtual_path = system_path
    if VIRTUAL_ENV:
        if virtual_path is None:
            virtual_path = []
        return os.path.join(*virtual_path)
    else:
        if system_path is None:
            system_path = []
        return os.path.join(*(["/"] + system_path))


extras_require = {}
for package in ["common", "client", "server"]:
    with open(os.path.join(package, "requirements.txt")) as f:
        extras_require[package] = f.readlines()

extras_require["all"] = list(
    set(
        requirement
        for requirements in extras_require.values()
        for requirement in requirements
    )
)

with open("test-requirements.txt") as f:
    test_requirements = f.readlines()

setup(
    name="odcs",
    description="On Demand Compose Service",
    version="0.2.51",
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Build Tools",
    ],
    keywords="on demand compose service modularity fedora",
    author="The Factory 2.0 Team",
    # TODO: Not sure which name would be used for mail alias,
    # but let's set this proactively to the new name.
    author_email="odcs-owner@fedoraproject.org",
    url="https://pagure.io/odcs/",
    license="GPLv2+",
    packages=["odcs", "odcs.client", "odcs.server", "odcs.common"],
    package_dir={
        "odcs": "common/odcs",
        "odcs.client": "client/odcs/client",
        "odcs.server": "server/odcs/server",
        "odcs.common": "common/odcs/common",
    },
    extras_require=extras_require,
    include_package_data=True,
    zip_safe=False,
    install_requires=extras_require["client"],
    tests_require=test_requirements,
    scripts=["client/contrib/odcs", "server/contrib/odcs-promote-compose"],
    entry_points={
        "console_scripts": [
            "odcs-upgradedb = odcs.server.manage:upgradedb [server]",
            "odcs-gencert = odcs.server.manage:generatelocalhostcert [server]",
            "odcs-frontend = odcs.server.manage:runssl [server]",
            "odcs-mock-runroot = odcs.server.mock_runroot:mock_runroot_main [server]",
            "odcs-manager = odcs.server.manage:manager_wrapper [server]",
        ],
    },
    data_files=[
        (
            get_dir(["etc", "odcs"]),
            [
                "server/conf/config.py",
                "server/conf/pungi.conf",
                "server/conf/raw_config_wrapper.conf",
            ],
        ),
    ],
)



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import odcs.server
from odcs.server import db, conf
from odcs.server.models import Compose
from odcs.common.types import COMPOSE_STATES, COMPOSE_RESULTS
from odcs.server.backend import RemoveExpiredComposesThread
from odcs.server.pungi import PungiSourceType
from datetime import datetime, timedelta

from .utils import ModelsBaseTest, AnyStringWith

import os
import mock
from mock import patch


class TestRemoveExpiredComposesThread(ModelsBaseTest):
    maxDiff = None

    def setUp(self):
        super(TestRemoveExpiredComposesThread, self).setUp()

        compose = Compose.create(
            db.session,
            "unknown",
            PungiSourceType.MODULE,
            "testmodule-master",
            COMPOSE_RESULTS["repository"],
            60,
        )
        db.session.add(compose)
        db.session.commit()

        self.thread = RemoveExpiredComposesThread()

    def test_does_not_remove_a_compose_which_state_is_not_done(self):
        """
        Test that we do not remove a composes on non-done state.
        """
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        c.time_to_expire = datetime.utcnow() - timedelta(seconds=-120)

        for name, state in COMPOSE_STATES.items():
            if name == "done":
                # Compose with state DONE would be removed.
                continue
            c.state = state
            db.session.add(c)
            db.session.commit()
            self.thread.do_work()
            db.session.expunge_all()
            c = db.session.query(Compose).filter(Compose.id == 1).one()
            self.assertEqual(c.state, state)

    @patch("os.unlink")
    @patch("os.path.realpath")
    @patch("os.path.exists")
    @patch("odcs.server.backend.get_latest_symlink")
    def test_a_compose_which_state_is_done_is_removed(
        self, latest_symlink, exists, realpath, unlink
    ):
        """
        Test that we do remove a compose in done state.
        """
        latest_symlink.return_value = None
        realpath.return_value = "/odcs-real"
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        c.time_to_expire = datetime.utcnow() - timedelta(seconds=120)
        c.state = COMPOSE_STATES["done"]
        c.compose_type = "nightly"
        c.pungi_compose_id = "compose-1-10-2020110.n.0"
        db.session.add(c)
        db.session.commit()
        self.thread.do_work()
        db.session.expunge_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["removed"])
        self.assertEqual(c.state_reason, "Compose is expired.")
        unlink.assert_has_calls(
            [
                mock.call(
                    AnyStringWith("test_composes/nightly/compose-1-10-2020110.n.0")
                ),
                mock.call(AnyStringWith("test_composes/odcs-1")),
            ]
        )

    @patch("os.unlink")
    @patch("os.path.realpath")
    @patch("os.path.exists")
    def test_latest_compose_not_removed(self, exists, realpath, unlink):
        """
        Test that we do remove a compose in done state.
        """
        realpath.return_value = "/odcs-real"
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        c.time_to_expire = datetime.utcnow() - timedelta(seconds=120)
        c.state = COMPOSE_STATES["done"]
        c.compose_type = "nightly"
        c.pungi_compose_id = "compose-1-10-2020110.n.0"
        db.session.add(c)
        db.session.commit()
        self.thread.do_work()
        db.session.expunge_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["done"])

    def test_a_compose_which_state_is_done_is_removed_keep_state_reason(self):
        """
        Test that we do remove a compose in done state.
        """
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        c.time_to_expire = datetime.utcnow() - timedelta(seconds=120)
        c.state = COMPOSE_STATES["done"]
        c.state_reason = "Generated successfully."
        db.session.add(c)
        db.session.commit()
        self.thread.do_work()
        db.session.expunge_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["removed"])
        self.assertEqual(c.state_reason, "Generated successfully.\nCompose is expired.")

    def test_does_not_remove_a_compose_which_is_not_expired(self):
        """
        Test that we do not remove a compose if its time_to_expire has not been
        reached yet.
        """
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        c.state = COMPOSE_STATES["done"]
        db.session.add(c)
        db.session.commit()
        self.thread.do_work()
        db.session.expunge_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["done"])

    def _mock_glob(self, glob, dirs):
        glob_ret_values = [[], []]
        for d in dirs:
            path = os.path.join(conf.target_dir, d)
            if d.startswith("latest-"):
                glob_ret_values[0].append(path)
            else:
                glob_ret_values[1].append(path)
        glob.side_effect = glob_ret_values

    @patch("os.path.isdir")
    @patch("glob.glob")
    @patch("odcs.server.backend.RemoveExpiredComposesThread._remove_compose_dir")
    def test_remove_left_composes(self, remove_compose_dir, glob, isdir):
        isdir.return_value = True
        self._mock_glob(glob, ["latest-odcs-96-1", "odcs-96-1-20171005.n.0", "odcs-96"])
        self.thread.do_work()
        self.assertEqual(
            remove_compose_dir.call_args_list,
            [
                mock.call(os.path.join(conf.target_dir, "latest-odcs-96-1")),
                mock.call(os.path.join(conf.target_dir, "odcs-96-1-20171005.n.0")),
                mock.call(os.path.join(conf.target_dir, "odcs-96")),
            ],
        )

    @patch("os.path.isdir")
    @patch("glob.glob")
    @patch("odcs.server.backend.RemoveExpiredComposesThread._remove_compose_dir")
    @patch.object(
        odcs.server.config.Config,
        "extra_target_dirs",
        new={"releng-private": "/tmp/private"},
    )
    def test_remove_left_composes_extra_target_dir(
        self, remove_compose_dir, glob, isdir
    ):
        isdir.return_value = True
        self.thread.do_work()
        print(glob.call_args_list)
        self.assertEqual(
            glob.call_args_list,
            [
                mock.call(os.path.join(conf.target_dir, "latest-odcs-*")),
                mock.call("/tmp/private/latest-odcs-*"),
                mock.call(os.path.join(conf.target_dir, "odcs-*")),
                mock.call("/tmp/private/odcs-*"),
            ],
        )

    @patch("os.path.isdir")
    @patch("glob.glob")
    @patch("odcs.server.backend.RemoveExpiredComposesThread._remove_compose_dir")
    def test_remove_left_composes_not_dir(self, remove_compose_dir, glob, isdir):
        isdir.return_value = False
        self._mock_glob(glob, ["latest-odcs-96-1"])
        self.thread.do_work()
        remove_compose_dir.assert_not_called()

    @patch("os.path.isdir")
    @patch("glob.glob")
    @patch("odcs.server.backend.RemoveExpiredComposesThread._remove_compose_dir")
    def test_remove_left_composes_wrong_dir(self, remove_compose_dir, glob, isdir):
        isdir.return_value = True
        self._mock_glob(glob, ["latest-odcs-", "odcs-", "odcs-abc"])
        self.thread.do_work()
        remove_compose_dir.assert_not_called()

    @patch("os.path.isdir")
    @patch("glob.glob")
    @patch("odcs.server.backend.RemoveExpiredComposesThread._remove_compose_dir")
    def test_remove_left_composes_valid_compose(self, remove_compose_dir, glob, isdir):
        isdir.return_value = True
        self._mock_glob(glob, ["latest-odcs-1-1", "odcs-1-1-2017.n.0"])
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        c.state = COMPOSE_STATES["done"]
        db.session.add(c)
        db.session.commit()
        self.thread.do_work()
        remove_compose_dir.assert_not_called()

    @patch("os.path.isdir")
    @patch("glob.glob")
    @patch("odcs.server.backend.RemoveExpiredComposesThread._remove_compose_dir")
    def test_remove_left_composes_expired_compose(
        self, remove_compose_dir, glob, isdir
    ):
        isdir.return_value = True
        self._mock_glob(glob, ["latest-odcs-1-1", "odcs-1-1-2017.n.0"])
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        c.state = COMPOSE_STATES["removed"]
        db.session.add(c)
        db.session.commit()
        self.thread.do_work()
        self.assertEqual(
            remove_compose_dir.call_args_list,
            [
                mock.call(os.path.join(conf.target_dir, "latest-odcs-1-1")),
                mock.call(os.path.join(conf.target_dir, "odcs-1-1-2017.n.0")),
            ],
        )

    @patch("shutil.rmtree")
    @patch("os.unlink")
    @patch("os.path.realpath")
    @patch("os.path.exists")
    def test_remove_compose_dir_symlink(self, exists, realpath, unlink, rmtree):
        exists.return_value = True
        toplevel_dir = "/odcs"
        realpath.return_value = "/odcs-real"

        self.thread._remove_compose_dir(toplevel_dir)
        unlink.assert_called_once()
        rmtree.assert_called_once()

    @patch("shutil.rmtree")
    @patch("os.unlink")
    @patch("os.path.realpath")
    @patch("os.path.exists")
    def test_remove_compose_dir_broken_symlink(self, exists, realpath, unlink, rmtree):
        def mocked_exists(p):
            return p != "/odcs-real"

        exists.side_effect = mocked_exists
        toplevel_dir = "/odcs"
        realpath.return_value = "/odcs-real"

        self.thread._remove_compose_dir(toplevel_dir)
        unlink.assert_called_once()
        rmtree.assert_not_called()

    @patch("shutil.rmtree")
    @patch("os.unlink")
    @patch("os.path.realpath")
    @patch("os.path.exists")
    def test_remove_compose_dir_real_dir(self, exists, realpath, unlink, rmtree):
        exists.return_value = True
        toplevel_dir = "/odcs"
        realpath.return_value = "/odcs"

        self.thread._remove_compose_dir(toplevel_dir)
        unlink.assert_not_called()
        rmtree.assert_called_once()

    @patch("os.unlink")
    @patch("os.path.realpath")
    @patch("os.path.exists")
    @patch("odcs.server.backend.log.warning")
    def test_remove_compose_rmtree_error(self, log_warning, exists, realpath, unlink):
        exists.return_value = True
        toplevel_dir = "/odcs"
        realpath.return_value = "/odcs-real"

        # This must not raise an exception.
        self.thread._remove_compose_dir(toplevel_dir)
        log_warning.assert_called_once_with(
            AnyStringWith("Cannot remove some files in /odcs-real:")
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

""" SQLAlchemy Database models for the Flask app
"""

import os

from datetime import datetime, timedelta

from flask_login import UserMixin
from sqlalchemy.orm import validates
from sqlalchemy.schema import Index

from odcs.server import conf
from odcs.server import db
from odcs.server.events import cache_composes_if_state_changed
from odcs.server.events import start_to_publish_messages
from odcs.common.types import (
    COMPOSE_STATES,
    INVERSE_COMPOSE_STATES,
    COMPOSE_FLAGS,
    COMPOSE_RESULTS,
    PungiSourceType,
)

from sqlalchemy import event, or_
from flask_sqlalchemy import SignallingSession

event.listen(SignallingSession, "after_flush", cache_composes_if_state_changed)

event.listen(SignallingSession, "after_commit", start_to_publish_messages)


def commit_on_success(func):
    def _decorator(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            db.session.rollback()
            raise
        finally:
            db.session.commit()

    return _decorator


class ODCSBase(db.Model):
    __abstract__ = True


class User(ODCSBase, UserMixin):
    """User information table"""

    __tablename__ = "users"

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(200), nullable=False, unique=True)

    @classmethod
    def find_user_by_name(cls, username):
        """Find a user by username

        :param str username: a string of username to find user
        :return: user object if found, otherwise None is returned.
        :rtype: User
        """
        try:
            return db.session.query(cls).filter(cls.username == username)[0]
        except IndexError:
            return None

    @classmethod
    def create_user(cls, username):
        user = cls(username=username)
        db.session.add(user)
        return user


class Compose(ODCSBase):
    __tablename__ = "composes"
    id = db.Column(db.Integer, primary_key=True)
    owner = db.Column(db.String, nullable=False)
    # PungiSourceType
    source_type = db.Column(db.Integer, nullable=False)
    # White-space separated list of koji_tags or modules
    source = db.Column(db.String, nullable=False)
    # White-space separated list of modular Koji tags.
    modular_koji_tags = db.Column(db.String)
    # URL on which the module defaults can be found.
    module_defaults_url = db.Column(db.String)
    # Koji event id at which the compose has been generated
    koji_event = db.Column(db.Integer)
    # White-space separated list sigkeys to define the key using which
    # the package in compose must be signed.
    sigkeys = db.Column(db.String)
    # COMPOSES_STATES
    state = db.Column(db.Integer, nullable=False, index=True)
    # Reason of state change
    state_reason = db.Column(db.String, nullable=True)
    # COMPOSE_RESULTS
    results = db.Column(db.Integer, nullable=False)
    # White-space separated list of packages
    packages = db.Column(db.String)
    # White-space separated list of builds (NVR) to include in a compose.
    builds = db.Column(db.String)
    # COMPOSE_FLAGS
    flags = db.Column(db.Integer)
    # Time when the compose should be deleted
    time_to_expire = db.Column(db.DateTime, nullable=False, index=True)
    # Time when the request for compose was submitted to ODCS
    time_submitted = db.Column(db.DateTime, nullable=False)
    # Time when compose transitioned from WAITING to GENERATING
    time_started = db.Column(db.DateTime)
    # Time when compose was successfully finished. Failed composes do not have
    # this time.
    time_done = db.Column(db.DateTime)
    # Time when compose was deleted
    time_removed = db.Column(db.DateTime)
    # removed_by is set when compose is deleted rather than expired normally
    removed_by = db.Column(db.String, nullable=True)
    reused_id = db.Column(db.Integer, index=True)
    # In case Pungi composes are generated using ODCS Koji runroot task, this
    # holds the Koji task id of this task.
    koji_task_id = db.Column(db.Integer, index=True)
    # White-space separated list of arches to build for.
    arches = db.Column(db.String)
    # White-space separated list of arches to enable multilib for.
    multilib_arches = db.Column(db.String)
    # Method to generate multilib compose as defined by python-multilib.
    multilib_method = db.Column(db.Integer)
    # White-space separated lookaside repository URLs.
    lookaside_repos = db.Column(db.String, nullable=True)
    # Compose label stored to ComposeInfo for Raw config composes.
    label = db.Column(db.String, nullable=True)
    # Compose type stored to ComposeInfo for Raw config composes.
    compose_type = db.Column(db.String, nullable=True)
    # Compose id as generated by Pungi for its ComposeInfo metadata.
    pungi_compose_id = db.Column(db.String, nullable=True)
    # Full Pungi configuration dump, used only for raw_config source type.
    pungi_config_dump = db.Column(db.String, nullable=True)
    # UUID of the celery task.
    celery_task_id = db.Column(db.String, nullable=True)
    # Target directory in which the compose is stored. This is `conf.target_dir`
    # by default.
    _target_dir = db.Column("target_dir", db.String, nullable=True)
    # White-space separated list of scratch modules (N:S:V:C) to include in a compose.
    scratch_modules = db.Column(db.String, nullable=True)
    # White-space separated list of RPM scratch builds to include in a compose.
    scratch_build_tasks = db.Column(db.String, nullable=True)
    # White-space separated list of parent Pungi compose IDs (pungi_compose_id).
    parent_pungi_compose_ids = db.Column(db.String, nullable=True)

    @property
    def on_default_target_dir(self):
        """
        True if this compose is stored on default `conf.target_dir`.
        """
        return self.target_dir is None or self.target_dir == conf.target_dir

    @property
    def target_dir(self):
        """
        Returns the `self._target_dir` if set, otherwise `conf.target_dir`.

        This is needed to keep backward compatibility with composes which do
        not have the `Compose.target_dir` set.
        """
        return self._target_dir or conf.target_dir

    @target_dir.setter
    def target_dir(self, value):
        self._target_dir = value

    @classmethod
    def create(
        cls,
        session,
        owner,
        source_type,
        source,
        results,
        seconds_to_live,
        packages=None,
        flags=0,
        sigkeys=None,
        koji_event=None,
        arches=None,
        multilib_arches=None,
        multilib_method=None,
        builds=None,
        lookaside_repos=None,
        modular_koji_tags=None,
        module_defaults_url=None,
        label=None,
        compose_type=None,
        target_dir=None,
        scratch_modules=None,
        parent_pungi_compose_ids=None,
        scratch_build_tasks=None,
    ):
        now = datetime.utcnow()
        compose = cls(
            owner=owner,
            source_type=source_type,
            source=source,
            sigkeys=sigkeys,
            koji_event=koji_event,
            state="wait",
            results=results,
            time_submitted=now,
            time_to_expire=now + timedelta(seconds=seconds_to_live),
            packages=packages,
            flags=flags,
            arches=arches if arches else " ".join(conf.arches),
            multilib_arches=multilib_arches if multilib_arches else "",
            multilib_method=multilib_method if multilib_method else 0,
            builds=builds,
            lookaside_repos=lookaside_repos,
            modular_koji_tags=modular_koji_tags,
            module_defaults_url=module_defaults_url,
            label=label,
            compose_type=compose_type,
            target_dir=target_dir or conf.target_dir,
            scratch_modules=scratch_modules,
            parent_pungi_compose_ids=parent_pungi_compose_ids,
            scratch_build_tasks=scratch_build_tasks,
        )
        session.add(compose)
        return compose

    @classmethod
    def create_copy(
        cls, session, compose, owner=None, seconds_to_live=None, sigkeys=None
    ):
        """
        Creates new compose with all the options influencing the resulting
        compose copied from the `compose`. The `owner` and `seconds_to_live`
        can be set independently. The state of copied compose is "wait".
        """
        now = datetime.utcnow()
        if not seconds_to_live:
            seconds_to_live = conf.seconds_to_live

        compose = cls(
            owner=owner or compose.owner,
            source_type=compose.source_type,
            source=compose.source,
            state="wait",
            results=compose.results,
            time_submitted=now,
            time_to_expire=now + timedelta(seconds=seconds_to_live),
            packages=compose.packages,
            builds=compose.builds,
            flags=compose.flags,
            koji_event=compose.koji_event,
            arches=compose.arches,
            multilib_arches=compose.multilib_arches,
            multilib_method=compose.multilib_method,
            sigkeys=sigkeys if sigkeys is not None else compose.sigkeys,
            lookaside_repos=compose.lookaside_repos,
            modular_koji_tags=compose.modular_koji_tags,
            module_defaults_url=compose.module_defaults_url,
            label=compose.label,
            compose_type=compose.compose_type,
            pungi_config_dump=compose.pungi_config_dump,
            # Set pungi_compose_id to None, because it is regenerated once
            # this copied Compose is started.
            pungi_compose_id=None,
            # Also reset celery task_id
            celery_task_id=None,
            target_dir=compose.target_dir,
            scratch_modules=compose.scratch_modules,
            parent_pungi_compose_ids=compose.parent_pungi_compose_ids,
            scratch_build_tasks=compose.scratch_build_tasks,
        )
        session.add(compose)
        return compose

    @property
    def name(self):
        if self.reused_id:
            return "odcs-%d" % self.reused_id
        else:
            return "odcs-%d" % self.id

    @property
    def toplevel_dir(self):
        return os.path.join(self.target_dir, self.name)

    @property
    def toplevel_url(self):
        if not self.on_default_target_dir:
            return ""

        return conf.target_dir_url + "/" + self.name

    @property
    def result_repo_dir(self):
        """
        Returns path to compose directory with per-arch repositories with
        results.
        """
        return os.path.join(self.toplevel_dir, "compose", "Temporary")

    @property
    def result_repo_url(self):
        """
        Returns public URL to compose directory with per-arch repositories.
        """
        if not self.on_default_target_dir:
            return ""

        return (
            conf.target_dir_url + "/" + os.path.join(self.name, "compose", "Temporary")
        )

    @property
    def result_repofile_path(self):
        """
        Returns path to .repo file.
        """
        return os.path.join(
            self.toplevel_dir, "compose", "Temporary", self.name + ".repo"
        )

    @property
    def result_repofile_url(self):
        """
        Returns public URL to repofile.
        """
        if not self.on_default_target_dir:
            return ""

        # There is no repofile for Raw config composes.
        if self.source_type == PungiSourceType.RAW_CONFIG:
            return ""

        return (
            conf.target_dir_url
            + "/"
            + os.path.join(self.name, "compose", "Temporary", self.name + ".repo")
        )

    @validates("state")
    def validate_state(self, key, field):
        if field in COMPOSE_STATES.values():
            return field
        if field in COMPOSE_STATES:
            return COMPOSE_STATES[field]
        raise ValueError("%s: %s, not in %r" % (key, field, COMPOSE_STATES))

    def json(self, full=False):
        flags = []
        for name, value in COMPOSE_FLAGS.items():
            if value == 0:
                continue
            if self.flags & value:
                flags.append(name)

        results = []
        for name, value in COMPOSE_RESULTS.items():
            if value == 0:
                continue
            if self.results & value:
                results.append(name)

        if self.on_default_target_dir:
            target_dir = "default"
        else:
            inverse_extra_target_dirs = {
                v: k for k, v in conf.extra_target_dirs.items()
            }
            target_dir = inverse_extra_target_dirs.get(self.target_dir, "unknown")

        ret = {
            "id": self.id,
            "owner": self.owner,
            "source_type": self.source_type,
            "source": self.source,
            "state": self.state,
            "state_name": INVERSE_COMPOSE_STATES[self.state],
            "state_reason": self.state_reason,
            "time_to_expire": self._utc_datetime_to_iso(self.time_to_expire),
            "time_submitted": self._utc_datetime_to_iso(self.time_submitted),
            "time_started": self._utc_datetime_to_iso(self.time_started),
            "time_done": self._utc_datetime_to_iso(self.time_done),
            "time_removed": self._utc_datetime_to_iso(self.time_removed),
            "removed_by": self.removed_by,
            "result_repo": self.result_repo_url,
            "result_repofile": self.result_repofile_url,
            "toplevel_url": self.toplevel_url,
            "flags": flags,
            "results": results,
            "sigkeys": self.sigkeys if self.sigkeys else "",
            "koji_event": self.koji_event,
            "koji_task_id": self.koji_task_id,
            "packages": self.packages,
            "builds": self.builds,
            "arches": self.arches,
            "multilib_arches": self.multilib_arches,
            "multilib_method": self.multilib_method,
            "lookaside_repos": self.lookaside_repos,
            "modular_koji_tags": self.modular_koji_tags,
            "module_defaults_url": self.module_defaults_url,
            "label": self.label,
            "compose_type": self.compose_type,
            "pungi_compose_id": self.pungi_compose_id,
            "target_dir": target_dir,
            "scratch_modules": self.scratch_modules,
            "parent_pungi_compose_ids": self.parent_pungi_compose_ids,
            "scratch_build_tasks": self.scratch_build_tasks,
        }

        if full:
            ret["pungi_config_dump"] = self.pungi_config_dump

        return ret

    @staticmethod
    def _utc_datetime_to_iso(datetime_object):
        """
        Takes a UTC datetime object and returns an ISO formatted string
        :param datetime_object: datetime.datetime
        :return: string with datetime in ISO format
        """
        if datetime_object:
            # Converts the datetime to ISO 8601
            return datetime_object.strftime("%Y-%m-%dT%H:%M:%SZ")

        return None

    @classmethod
    def composes_to_expire(cls):
        now = datetime.utcnow()
        return Compose.query.filter(
            or_(
                Compose.state == COMPOSE_STATES["done"],
                Compose.state == COMPOSE_STATES["failed"],
            ),
            Compose.time_to_expire < now,
        ).all()

    def __repr__(self):
        return "<Compose %r, type %r, state %s>" % (
            self.id,
            self.source_type,
            INVERSE_COMPOSE_STATES[self.state],
        )

    def get_reused_compose(self):
        """Get compose this compose reuses"""
        return db.session.query(Compose).filter(Compose.id == self.reused_id).first()

    def get_reusing_composes(self):
        """Get composes that are reusing this compose"""
        return db.session.query(Compose).filter(Compose.reused_id == self.id).all()

    def extend_expiration(self, _from, seconds_to_live):
        """Extend time to expire"""
        new_expiration = max(
            self.time_to_expire, _from + timedelta(seconds=seconds_to_live)
        )
        if new_expiration != self.time_to_expire:
            self.time_to_expire = new_expiration

    def transition(self, to_state, reason, happen_on=None):
        """Transit compose state to a new state

        :param str to_state: transit this compose state to this state.
        :param str reason: the reason of this transition.
        :param happen_on: when this transition happens. Default is utcnow.
        :type happen_on: DateTime
        """
        self.state = to_state
        self.state_reason = reason
        if to_state == COMPOSE_STATES["removed"]:
            self.time_removed = happen_on or datetime.utcnow()
        elif to_state == COMPOSE_STATES["done"]:
            self.time_done = happen_on or datetime.utcnow()
        elif to_state == COMPOSE_STATES["generating"]:
            self.time_started = happen_on or datetime.utcnow()
        if to_state in (COMPOSE_STATES["done"], COMPOSE_STATES["failed"]):
            ttl = self.time_to_expire - self.time_submitted
            self.time_to_expire = (happen_on or datetime.utcnow()) + ttl
        db.session.commit()


Index("idx_source_type__state", Compose.source_type, Compose.state)



"""Add target_dir column.

Revision ID: 812f2745248f
Revises: a855c39e2a0f
Create Date: 2020-03-13 13:28:24.381076

"""

# revision identifiers, used by Alembic.
revision = "812f2745248f"
down_revision = "a855c39e2a0f"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("composes", sa.Column("target_dir", sa.String(), nullable=True))


def downgrade():
    op.drop_column("composes", "target_dir")



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import json
import os
import time
from datetime import datetime, timedelta

from mock import patch, MagicMock

import odcs.server
from odcs.server import db, app, conf
from odcs.server.models import Compose
from odcs.common.types import COMPOSE_STATES, COMPOSE_RESULTS, COMPOSE_FLAGS
from odcs.server.backend import ComposerThread, resolve_compose
from odcs.server.pungi import PungiSourceType

from .utils import ModelsBaseTest
from .mbs import mock_mbs

thisdir = os.path.abspath(os.path.dirname(__file__))


class TestComposerThread(ModelsBaseTest):
    maxDiff = None

    def setUp(self):
        self.client = app.test_client()
        super(TestComposerThread, self).setUp()
        self.composer = ComposerThread()

        patched_pungi_conf_path = os.path.join(thisdir, "../conf/pungi.conf")
        self.patch_pungi_conf_path = patch.object(
            odcs.server.conf, "pungi_conf_path", new=patched_pungi_conf_path
        )
        self.patch_pungi_conf_path.start()

        self.patch_update_cache = patch("odcs.server.backend.KojiTagCache.update_cache")
        self.update_cache = self.patch_update_cache.start()

    def tearDown(self):
        super(TestComposerThread, self).tearDown()
        self.patch_pungi_conf_path.stop()
        self.patch_update_cache.stop()

    def _wait_for_compose_state(self, id, state):
        c = None
        for i in range(20):
            db.session.expire_all()
            c = db.session.query(Compose).filter(Compose.id == id).one()
            if c.state == state:
                return c
            time.sleep(0.1)
        return c

    def _add_module_compose(self, source="testmodule-master-20170515074419", flags=0):
        compose = Compose.create(
            db.session,
            "unknown",
            PungiSourceType.MODULE,
            source,
            COMPOSE_RESULTS["repository"],
            60,
        )
        db.session.add(compose)
        db.session.commit()

    def _add_tag_compose(self, packages=None, flags=0):
        compose = Compose.create(
            db.session,
            "unknown",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages,
            flags,
        )
        db.session.add(compose)
        db.session.commit()

    def _add_repo_composes(self):
        old_c = Compose.create(
            db.session,
            "me",
            PungiSourceType.REPO,
            os.path.join(thisdir, "repo"),
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        old_c.state = COMPOSE_STATES["done"]
        resolve_compose(old_c)
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.REPO,
            os.path.join(thisdir, "repo"),
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        db.session.add(old_c)
        db.session.add(c)
        db.session.commit()

    @mock_mbs
    @patch("odcs.server.utils.execute_cmd")
    @patch("odcs.server.backend._write_repo_file")
    def test_submit_build(self, wrf, execute_cmd):
        self._add_module_compose()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

        self.assertEqual(self.composer.currently_generating, [])

        self.composer.do_work()
        c = self._wait_for_compose_state(1, COMPOSE_STATES["done"])
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(
                odcs.server.conf.target_dir, "latest-odcs-1-1/compose/Temporary"
            ),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/latest-odcs-1-1/compose/Temporary"
        )
        self.assertEqual(self.composer.currently_generating, [1])

    @mock_mbs
    @patch("odcs.server.utils.execute_cmd")
    @patch("odcs.server.backend._write_repo_file")
    def test_submit_build_module_without_release(self, wrf, execute_cmd):
        self._add_module_compose("testmodule-master")
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

        self.composer.do_work()
        c = self._wait_for_compose_state(1, COMPOSE_STATES["done"])
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(
                odcs.server.conf.target_dir, "latest-odcs-1-1/compose/Temporary"
            ),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/latest-odcs-1-1/compose/Temporary"
        )
        self.assertEqual(c.source, "testmodule:master:20170515074419")

    @mock_mbs
    @patch("odcs.server.utils.execute_cmd")
    @patch("odcs.server.backend._write_repo_file")
    def test_submit_build_colon_separator(self, wrf, execute_cmd):
        self._add_module_compose("testmodule:master:20170515074419")
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

        self.assertEqual(self.composer.currently_generating, [])

        self.composer.do_work()
        c = self._wait_for_compose_state(1, COMPOSE_STATES["done"])
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(
                odcs.server.conf.target_dir, "latest-odcs-1-1/compose/Temporary"
            ),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/latest-odcs-1-1/compose/Temporary"
        )
        self.assertEqual(self.composer.currently_generating, [1])

    @mock_mbs
    @patch("odcs.server.utils.execute_cmd")
    @patch("odcs.server.backend._write_repo_file")
    def test_submit_build_module_without_release_colon_separator(
        self, wrf, execute_cmd
    ):
        self._add_module_compose("testmodule:master")
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

        self.composer.do_work()
        c = self._wait_for_compose_state(1, COMPOSE_STATES["done"])
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(
                odcs.server.conf.target_dir, "latest-odcs-1-1/compose/Temporary"
            ),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/latest-odcs-1-1/compose/Temporary"
        )
        self.assertEqual(c.source, "testmodule:master:20170515074419")

    @mock_mbs
    @patch("odcs.server.utils.execute_cmd")
    @patch("odcs.server.backend._write_repo_file")
    def test_submit_build_module_without_release_not_in_mbs(self, wrf, execute_cmd):

        self._add_module_compose("testmodule2-master")
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

        self.composer.do_work()
        c = self._wait_for_compose_state(1, COMPOSE_STATES["failed"])
        self.assertEqual(c.state, COMPOSE_STATES["failed"])

    @mock_mbs
    @patch("odcs.server.backend.validate_pungi_compose")
    def test_submit_build_reuse_repo(self, mock_validate_pungi_compose):
        self._add_repo_composes()
        c = db.session.query(Compose).filter(Compose.id == 2).one()

        self.composer.do_work()
        c = self._wait_for_compose_state(2, COMPOSE_STATES["done"])
        self.assertEqual(c.reused_id, 1)
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(
                odcs.server.conf.target_dir, "latest-odcs-1-1/compose/Temporary"
            ),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/latest-odcs-1-1/compose/Temporary"
        )
        mock_validate_pungi_compose.assert_called_once()

    @mock_mbs
    def test_submit_build_reuse_module(self):
        self._add_module_compose()
        self._add_module_compose()
        old_c = db.session.query(Compose).filter(Compose.id == 1).one()
        old_c.state = COMPOSE_STATES["done"]
        resolve_compose(old_c)
        db.session.commit()

        self.composer.do_work()
        c = self._wait_for_compose_state(2, COMPOSE_STATES["done"])
        self.assertEqual(c.reused_id, 1)
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(
                odcs.server.conf.target_dir, "latest-odcs-1-1/compose/Temporary"
            ),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/latest-odcs-1-1/compose/Temporary"
        )

    @mock_mbs
    @patch("odcs.server.utils.execute_cmd")
    @patch("odcs.server.backend._write_repo_file")
    def test_submit_build_no_reuse_module(self, wrf, execute_cmd):
        self._add_module_compose()
        self._add_module_compose("testmodule-master-20170515074418")
        old_c = db.session.query(Compose).filter(Compose.id == 1).one()
        old_c.state = COMPOSE_STATES["done"]
        resolve_compose(old_c)
        db.session.commit()

        self.composer.do_work()
        c = self._wait_for_compose_state(2, COMPOSE_STATES["done"])
        self.assertEqual(c.reused_id, None)
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(
                odcs.server.conf.target_dir, "latest-odcs-2-1/compose/Temporary"
            ),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/latest-odcs-2-1/compose/Temporary"
        )

    @patch("odcs.server.backend.create_koji_session")
    @patch("odcs.server.backend._write_repo_file")
    def test_submit_build_no_deps(self, wrf, create_koji_session):
        """
        Checks that "no_deps" flags properly sets gather_method to nodeps.
        """
        koji_session = MagicMock()
        create_koji_session.return_value = koji_session
        koji_session.getLastEvent.return_value = {"id": 123}

        def mocked_execute_cmd(args, stdout=None, stderr=None, cwd=None, **kwargs):
            pungi_cfg = open(os.path.join(cwd, "pungi.json"), "r").read()
            pungi_cfg = json.loads(pungi_cfg)
            self.assertEqual(pungi_cfg["gather_method"], "nodeps")

        with patch("odcs.server.utils.execute_cmd", new=mocked_execute_cmd):
            self._add_tag_compose(flags=COMPOSE_FLAGS["no_deps"])
            c = db.session.query(Compose).filter(Compose.id == 1).one()
            self.assertEqual(c.state, COMPOSE_STATES["wait"])

            self.composer.do_work()
            c = self._wait_for_compose_state(1, COMPOSE_STATES["done"])
            self.assertEqual(c.state, COMPOSE_STATES["done"])

    @patch("odcs.server.backend.create_koji_session")
    def test_submit_build_reuse_koji_tag(self, create_koji_session):
        koji_session = MagicMock()
        create_koji_session.return_value = koji_session
        koji_session.getLastEvent.return_value = {"id": 123}
        koji_session.tagChangedSinceEvent.return_value = False

        self._add_tag_compose()
        self._add_tag_compose()
        old_c = db.session.query(Compose).filter(Compose.id == 1).one()
        old_c.state = COMPOSE_STATES["done"]
        resolve_compose(old_c)
        db.session.commit()

        self.composer.do_work()
        c = self._wait_for_compose_state(2, COMPOSE_STATES["done"])
        self.assertEqual(c.reused_id, 1)
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(odcs.server.conf.target_dir, "odcs-1/compose/Temporary"),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/odcs-1/compose/Temporary"
        )

    @patch("odcs.server.utils.execute_cmd")
    @patch("odcs.server.backend.create_koji_session")
    @patch("odcs.server.backend._write_repo_file")
    def test_submit_build_reuse_koji_tag_tags_changed(
        self, wrf, create_koji_session, execute_cmd
    ):
        koji_session = MagicMock()
        create_koji_session.return_value = koji_session
        koji_session.getLastEvent.return_value = {"id": 123}
        koji_session.tagChangedSinceEvent.return_value = True

        self._add_tag_compose()
        self._add_tag_compose()
        old_c = db.session.query(Compose).filter(Compose.id == 1).one()
        old_c.state = COMPOSE_STATES["done"]
        resolve_compose(old_c)
        db.session.commit()

        self.composer.do_work()
        c = self._wait_for_compose_state(2, COMPOSE_STATES["done"])
        self.assertEqual(c.reused_id, None)
        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(
            c.result_repo_dir,
            os.path.join(odcs.server.conf.target_dir, "odcs-2/compose/Temporary"),
        )
        self.assertEqual(
            c.result_repo_url, "http://localhost/odcs/odcs-2/compose/Temporary"
        )


class TestComposerThreadLostComposes(ModelsBaseTest):
    maxDiff = None

    def setUp(self):
        self.client = app.test_client()
        super(TestComposerThreadLostComposes, self).setUp()
        self.composer = ComposerThread()

        self.patch_generate_new_compose = patch(
            "odcs.server.backend.ComposerThread.generate_new_compose"
        )
        self.generate_new_compose = self.patch_generate_new_compose.start()

    def tearDown(self):
        super(TestComposerThreadLostComposes, self).tearDown()
        self.patch_generate_new_compose.stop()

    def _add_test_compose(self, state):
        compose = Compose.create(
            db.session,
            "unknown",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            "",
            0,
        )
        compose.state = state
        db.session.add(compose)
        db.session.commit()
        return compose

    def test_generate_lost_composes_generating_state(self):
        compose = self._add_test_compose(COMPOSE_STATES["generating"])
        self.composer.do_work()
        self.generate_new_compose.assert_called_once_with(compose)

    def test_generate_lost_composes_currently_generating(self):
        compose = self._add_test_compose(COMPOSE_STATES["generating"])
        self.composer.currently_generating.append(compose.id)
        self.composer.do_work()
        self.generate_new_compose.assert_not_called()

    def test_generate_lost_composes_all_states(self):
        for state in ["wait", "done", "removed", "failed"]:
            self._add_test_compose(COMPOSE_STATES[state])

        self.composer.generate_lost_composes()
        self.generate_new_compose.assert_not_called()

    def test_refresh_currently_generating(self):
        generating = self._add_test_compose(COMPOSE_STATES["generating"])
        done = self._add_test_compose(COMPOSE_STATES["done"])

        self.composer.currently_generating += [done.id, generating.id]
        self.composer.do_work()
        self.assertEqual(self.composer.currently_generating, [generating.id])


class TestComposerThreadStuckWaitComposes(ModelsBaseTest):
    maxDiff = None

    def setUp(self):
        self.client = app.test_client()
        super(TestComposerThreadStuckWaitComposes, self).setUp()
        self.composer = ComposerThread()

        self.patch_generate_new_compose = patch(
            "odcs.server.backend.ComposerThread.generate_new_compose"
        )
        self.generate_new_compose = self.patch_generate_new_compose.start()

    def tearDown(self):
        super(TestComposerThreadStuckWaitComposes, self).tearDown()
        self.patch_generate_new_compose.stop()

    def _add_test_compose(
        self,
        state,
        time_submitted=None,
        time_started=None,
        source_type=PungiSourceType.KOJI_TAG,
    ):
        compose = Compose.create(
            db.session,
            "unknown",
            source_type,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            "",
            0,
        )
        compose.state = state
        if time_submitted:
            compose.time_submitted = time_submitted
        if time_started:
            compose.time_started = time_started
        db.session.add(compose)
        db.session.commit()
        return compose

    def test_fail_lost_generating_composes(self):
        t = datetime.utcnow() - timedelta(seconds=2 * conf.pungi_timeout)

        time_submitted = t - timedelta(minutes=6)
        time_started = t - timedelta(minutes=5)
        compose_to_fail = self._add_test_compose(
            COMPOSE_STATES["generating"],
            time_submitted=time_submitted,
            time_started=time_started,
        )

        time_submitted = t + timedelta(minutes=4)
        time_started = t + timedelta(minutes=5)
        compose_to_keep = self._add_test_compose(
            COMPOSE_STATES["generating"],
            time_submitted=time_submitted,
            time_started=time_started,
        )

        self.composer.fail_lost_generating_composes()
        db.session.commit()
        db.session.expire_all()

        self.assertEqual(compose_to_fail.state, COMPOSE_STATES["failed"])
        self.assertEqual(compose_to_keep.state, COMPOSE_STATES["generating"])



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import errno
import koji
import os
import threading
import shutil
import six
import productmd.compose
import productmd.common
from datetime import datetime, timedelta
from odcs.server import log, conf, app, db
from odcs.server.models import Compose, COMPOSE_STATES, COMPOSE_FLAGS
from odcs.server.pungi import (
    Pungi,
    PungiConfig,
    PungiSourceType,
    PungiLogs,
    RawPungiConfig,
)
from odcs.server.pulp import Pulp
from odcs.server.cache import KojiTagCache
from odcs.server.pungi_compose import PungiCompose
from concurrent.futures import ThreadPoolExecutor
import glob
import odcs.server.utils
import odcs.server.mbs
import defusedxml.ElementTree

# Cache last event for each tag since that a compose was generated from that
# tag.
# This is a mapping from tag name to koji event id. For example,
# {
#     'tag1': 123456,
#     'tag2': 123476,
# }
LAST_EVENTS_CACHE = {}


class BackendThread(object):
    """
    Base BackendThread class.

    The `BackendThread.do_work(...)` is called repeatedly after `timeout`
    seconds.
    """

    def __init__(self, timeout=1):
        """
        Creates new BackendThread instance.

        :param int timeout: Timeout in seconds after which do_work is called.
        """
        self.thread = None
        self.exit = False
        self.exit_cond = threading.Condition()
        self.timeout = timeout

    def do_work(self):
        """
        Reimplement this method in your own BackendThread subclass.
        This method is called every `timeout` seconds.
        """
        raise NotImplementedError("do_work() method not implemented")

    def _run(self):
        """
        Main "run" method of a thread. Calls `do_work()` after `self.timeout`
        seconds. Stops then `stop()` is called.
        """
        while not self.exit:
            try:
                self.do_work()
            except Exception:
                log.exception("Exception in backend thread")
                try:
                    db.session.rollback()
                except Exception:
                    log.exception("Cannot rollback DB session")

            # If the self.stop has been called from the do_work, break earlier
            # so we don't wait on exit_cond.
            if self.exit:
                break

            self.exit_cond.acquire()
            self.exit_cond.wait(float(self.timeout))
            self.exit_cond.release()

    def join(self):
        """
        Waits until the thread terminates.
        """
        self.thread.join()

    def stop(self):
        """
        Stops the thread.
        """
        self.exit = True
        self.exit_cond.acquire()
        self.exit_cond.notify()
        self.exit_cond.release()

    def start(self):
        """
        Starts the thread.
        """
        self.thread = threading.Thread(target=self._run)
        self.thread.setDaemon(True)
        self.thread.start()


class RemoveExpiredComposesThread(BackendThread):
    """
    Thread used to remove old expired composes.
    """

    def __init__(self):
        """
        Creates new RemoveExpiredComposesThread instance.
        """
        super(RemoveExpiredComposesThread, self).__init__(10)
        self._rmtree_errors = {}

    def _on_rmtree_error(self, function, path, excinf):
        """
        Helper method passed to `shutil.rmtree` as `onerror` kwarg which stores
        the rmtree errors in the `self._rmtree_errors` dict and allows the rmtree
        to continue removing other files.

        The errors stored in `self._rmtree_errors` are grouped by the
        `dirname(path)`. This is needed to not print thousands of log lines in
        case when directory with wrong permissions contain thousands of files.
        """
        if function == os.remove:
            d = os.path.dirname(path)
        else:
            d = path
        if d in self._rmtree_errors:
            # This directory has been already handled.
            return
        exception_value = excinf[1]
        self._rmtree_errors[d] = repr(exception_value)

    def _remove_compose_dir(self, toplevel_dir):
        """
        Removes the compose toplevel_dir symlink together with the real
        path it points to.
        """

        # Be nice and don't fail when directory does not exist.
        if not os.path.exists(toplevel_dir):
            log.warning("Cannot remove directory %s, it does not exist", toplevel_dir)
            return

        # Temporary dictionary to store errors from `self._on_rmtree_error`.
        self._rmtree_errors = {}

        # If toplevel_dir is a symlink, remove the symlink and
        # its target. If toplevel_dir is normal directory, just
        # remove it using rmtree.
        if os.path.realpath(toplevel_dir) != toplevel_dir:
            targetpath = os.path.realpath(toplevel_dir)
            os.unlink(toplevel_dir)
            if os.path.exists(targetpath):
                shutil.rmtree(targetpath, onerror=self._on_rmtree_error)
        else:
            shutil.rmtree(toplevel_dir, onerror=self._on_rmtree_error)

        for path, error in self._rmtree_errors.items():
            log.warning("Cannot remove some files in %s: %r" % (path, error))

    def _get_compose_id_from_path(self, path):
        """
        Returns the ID of compose from directory path in conf.target_dir.
        """
        parts = os.path.basename(path).split("-")
        while parts and parts[0] != "odcs":
            del parts[0]

        if not parts or len(parts) < 2 or not parts[1].isdigit():
            log.error("Directory %s is not valid compose directory", path)
            return None

        return int(parts[1])

    def do_work(self):
        """
        Checks for the expired composes and removes them.
        """
        log.info("Checking for expired composes")

        composes = Compose.composes_to_expire()
        for compose in composes:
            # Check if target_dir is accessible on this backend and if not, skip it.
            if not os.path.exists(compose.target_dir):
                continue

            # Do not remove the latest successfull compose.
            if get_latest_symlink(compose):
                log.info(
                    "%r: Not removing compose, the latest-* symlink points to it.",
                    compose,
                )
                continue

            log.info("%r: Removing compose", compose)
            if compose.removed_by:
                state_reason = "Removed by {}.".format(compose.removed_by)
            else:
                state_reason = "Compose is expired."
            if compose.state_reason:
                state_reason = "{}\n{}".format(compose.state_reason, state_reason)
            compose.transition(COMPOSE_STATES["removed"], state_reason)
            if not compose.reused_id:
                self._remove_compose_dir(compose.toplevel_dir)
                remove_compose_symlink(compose)

                # Remove the `odcs-$COMPOSE_ID` symlink if it's there.
                symlink = os.path.join(compose.target_dir, compose.name)
                if os.path.exists(symlink) and os.path.realpath(symlink) != symlink:
                    os.unlink(symlink)

        # In case of ODCS error, there might be left-over directories
        # belonging to already expired composes. Try to find them in the
        # target_dir.
        # At first, get all the directories in target_dir which are created
        # by ODCS.
        odcs_paths = []
        for dirname in ["latest-odcs-*", "odcs-*"]:
            for target_dir in [conf.target_dir] + list(conf.extra_target_dirs.values()):
                path = os.path.join(target_dir, dirname)
                odcs_paths += glob.glob(path)

        # Then try removing them if they are left there by some error.
        for path in odcs_paths:
            # Check that we are really going to remove a directory.
            if not os.path.isdir(path):
                continue

            compose_id = self._get_compose_id_from_path(path)
            if not compose_id:
                # Error logged in _get_compose_id_from_dirname already.
                continue

            composes = Compose.query.filter(Compose.id == compose_id).all()
            if not composes:
                log.info(
                    "Removing data of compose %d - it is not in " "database: %s",
                    compose_id,
                    path,
                )
                self._remove_compose_dir(path)
                continue

            compose = composes[0]
            if compose.state == COMPOSE_STATES["removed"]:
                log.info(
                    "%r: Removing data of compose - it has already "
                    "expired some time ago: %s",
                    compose_id,
                    path,
                )
                self._remove_compose_dir(path)
                continue

        # Remove old Koji tag data from Koji tag cache.
        KojiTagCache.remove_old_koji_tag_cache_data()


def create_koji_session():
    """
    Creates and returns new koji_session based on the `conf.koji_profile`.
    """

    koji_module = koji.get_profile_module(conf.koji_profile)
    session_opts = {}
    for key in (
        "krbservice",
        "timeout",
        "keepalive",
        "max_retries",
        "retry_interval",
        "anon_retry",
        "offline_retry",
        "offline_retry_interval",
        "debug",
        "debug_xmlrpc",
        "krb_rdns",
        "use_fast_upload",
    ):
        value = getattr(koji_module.config, key, None)
        if value is not None:
            session_opts[key] = value
    koji_session = koji.ClientSession(koji_module.config.server, session_opts)
    return koji_session


def koji_get_inherited_tags(koji_session, tag, tags=None):
    """
    Returns list of ids of all tags the tag `tag` inherits from.
    """

    info = koji_session.getTag(tag)
    if not info:
        raise ValueError("Unknown Koji tag %s." % tag)
    ids = [info["id"]]
    seen_tags = tags or set()
    inheritance_data = koji_session.getInheritanceData(tag)
    inheritance_data = [
        data for data in inheritance_data if data["parent_id"] not in seen_tags
    ]

    # Iterate over all the tags this tag inherits from.
    for inherited in inheritance_data:
        # Make a note to ourselves that we have seen this parent_tag.
        parent_tag_id = inherited["parent_id"]
        seen_tags.add(parent_tag_id)

        # Get tag info for the parent_tag.
        info = koji_session.getTag(parent_tag_id)
        if info is None:
            log.error("Cannot get info about Koji tag %s", parent_tag_id)
            return []

        ids += koji_get_inherited_tags(koji_session, info["name"], seen_tags)

    return ids


def tag_changed(koji_session, tag, koji_event):
    """
    Check if tag and its parents in the inheritance have changed since a
    particular koji event

    :param koji_session: instance of :class:`ClientSession`.
    :param str tag: tag name.
    :param int koji_event: the koji event id.
    :return: True if changed, otherwise False.
    :rtype: bool
    """
    tags = koji_get_inherited_tags(koji_session, tag)
    return koji_session.tagChangedSinceEvent(koji_event, tags)


def resolve_compose(compose):
    """
    Resolves various general compose values to the real ones. For example:
    - Sets the koji_event based on the current Koji event, so it can be used
      to generate the compose and we can find out if we can reuse that compose
      later
    - For MODULE PungiSourceType, resolves the modules without the "release"
      field to latest module release using MBS.
    """
    if compose.source_type == PungiSourceType.REPO:
        # We treat "revision" of local repo as koji_event for the simplicity.
        repomd = os.path.join(compose.source, "repodata", "repomd.xml")
        e = defusedxml.ElementTree.parse(repomd).getroot()
        revision = e.find("{http://linux.duke.edu/metadata/repo}revision").text
        compose.koji_event = int(revision)
    elif compose.source_type == PungiSourceType.KOJI_TAG:
        global LAST_EVENTS_CACHE
        koji_session = create_koji_session()
        # If compose.koji_event is set, it means that we are regenerating
        # previous compose and we have to respect the previous koji_event to
        # get the same results.
        if not compose.koji_event:
            if compose.source not in LAST_EVENTS_CACHE:
                event_id = int(koji_session.getLastEvent()["id"])
            elif tag_changed(
                koji_session, compose.source, LAST_EVENTS_CACHE[compose.source]
            ):
                event_id = int(koji_session.getLastEvent()["id"])
            else:
                event_id = LAST_EVENTS_CACHE[compose.source]
                log.info(
                    "Reuse koji event %s to generate compose %s from source %s",
                    event_id,
                    compose.id,
                    compose.source,
                )
            compose.koji_event = event_id
            # event_id could be a new koji event ID. Cache it for next potential
            # reuse for same tag.
            LAST_EVENTS_CACHE[compose.source] = event_id
    elif compose.source_type == PungiSourceType.MODULE:

        # Resolve the latest release of modules which do not have the release
        # string defined in the compose.source.
        mbs = odcs.server.mbs.MBS(conf)
        modules = compose.source.split(" ")

        include_done = compose.flags & COMPOSE_FLAGS["include_done_modules"]
        specified_mbs_modules = []
        for module in modules:
            # In case the module is defined by complete NSVC, include it in a compose
            # even if it is in "done" state, because submitter directly asked for this
            # NSVC.
            is_complete_nsvc = module.count(":") == 3
            specified_mbs_modules += mbs.get_latest_modules(
                module, include_done or is_complete_nsvc
            )

        expand = not compose.flags & COMPOSE_FLAGS["no_deps"]
        new_mbs_modules = mbs.validate_module_list(specified_mbs_modules, expand=expand)

        uids = sorted(
            "{name}:{stream}:{version}:{context}".format(**m)
            for m in new_mbs_modules
            if m["name"] not in conf.base_module_names
        )
        compose.source = " ".join(uids)
    elif compose.source_type == PungiSourceType.PUNGI_COMPOSE:
        external_compose = PungiCompose(compose.source)
        rpms_data = external_compose.get_rpms_data()

        # If there is None in the sigkeys, it means unsigned packages are
        # allowed. The sigkeys in the `compose.sigkeys` are sorted by
        # preference and unsigned packages should be tried as last.
        # Therefore we need to remove None from `sigkeys` and handle
        # it as last element in `compose.sigkeys`.
        if None in rpms_data["sigkeys"]:
            allow_unsigned = True
            # Remove None from sigkeys.
            rpms_data["sigkeys"].remove(None)
        else:
            allow_unsigned = False
        compose.sigkeys = " ".join(rpms_data["sigkeys"])
        if allow_unsigned:
            # Unsigned packages are allowed by white-space in the end of
            # `compose.sigkeys`.
            compose.sigkeys += " "

        compose.arches = " ".join(rpms_data["arches"])
        compose.builds = " ".join(rpms_data["builds"].keys())

        packages = set()
        for rpms in rpms_data["builds"].values():
            for rpm_nevra in rpms:
                packages.add(productmd.common.parse_nvra(rpm_nevra)["name"])
        compose.packages = " ".join(packages)


def get_reusable_compose(compose):
    """
    Returns the compose in the "done" state which contains the same artifacts
    and results as the compose `compose` and therefore could be reused instead
    of generating new one.

    :param models.Compose compose: Instance of models.Compose.
    """

    if compose.flags & COMPOSE_FLAGS["no_reuse"]:
        return None

    # RAW_CONFIG composes cannot reuse other composes, we cannot track input
    # for them.
    if compose.source_type == PungiSourceType.RAW_CONFIG:
        return None

    # Get all the active composes of the same source_type
    composes = (
        db.session.query(Compose)
        .filter(
            Compose.state == COMPOSE_STATES["done"],
            Compose.source_type == compose.source_type,
        )
        .all()
    )

    for old_compose in composes:
        # Skip the old_compose in case it reuses another compose. In that case
        # the reused compose is also in composes list, so we won't miss it. We
        # don't want chain of composes reusing each other, because it would
        # break the time_to_expire handling.
        if old_compose.reused_id:
            continue

        packages = set(compose.packages.split(" ")) if compose.packages else set()
        old_packages = (
            set(old_compose.packages.split(" ")) if old_compose.packages else set()
        )
        if packages != old_packages:
            log.debug("%r: Cannot reuse %r - packages not same", compose, old_compose)
            continue

        builds = set(compose.builds.split(" ")) if compose.builds else set()
        old_builds = set(old_compose.builds.split(" ")) if old_compose.builds else set()
        if builds != old_builds:
            log.debug("%r: Cannot reuse %r - builds not same", compose, old_compose)
            continue

        source = set(compose.source.split(" "))
        old_source = set(old_compose.source.split(" "))
        if source != old_source:
            log.debug("%r: Cannot reuse %r - sources not same", compose, old_compose)
            continue

        if compose.flags != old_compose.flags:
            log.debug(
                "%r: Cannot reuse %r - flags not same, %d != %d",
                compose,
                old_compose,
                compose.flags,
                old_compose.flags,
            )
            continue

        if compose.results != old_compose.results:
            log.debug(
                "%r: Cannot reuse %r - results not same, %d != %d",
                compose,
                old_compose,
                compose.results,
                old_compose.results,
            )
            continue

        sigkeys = set(compose.sigkeys.split(" ")) if compose.sigkeys else set()
        old_sigkeys = (
            set(old_compose.sigkeys.split(" ")) if old_compose.sigkeys else set()
        )
        if sigkeys != old_sigkeys:
            log.debug("%r: Cannot reuse %r - sigkeys not same", compose, old_compose)
            continue

        arches = set(compose.arches.split(" ")) if compose.arches else set()
        old_arches = set(old_compose.arches.split(" ")) if old_compose.arches else set()
        if arches != old_arches:
            log.debug("%r: Cannot reuse %r - arches not same", compose, old_compose)
            continue

        lookaside_repos = (
            set(compose.lookaside_repos.split(" "))
            if compose.lookaside_repos
            else set()
        )
        old_lookaside_repos = (
            set(old_compose.lookaside_repos.split(" "))
            if old_compose.lookaside_repos
            else set()
        )
        if lookaside_repos != old_lookaside_repos:
            log.debug(
                "%r: Cannot reuse %r - lookaside_repos not same", compose, old_compose
            )
            continue

        multilib_arches = (
            set(compose.multilib_arches.split(" "))
            if compose.multilib_arches
            else set()
        )
        old_multilib_arches = (
            set(old_compose.multilib_arches.split(" "))
            if old_compose.multilib_arches
            else set()
        )
        if multilib_arches != old_multilib_arches:
            log.debug(
                "%r: Cannot reuse %r - multilib_arches not same", compose, old_compose
            )
            continue

        multilib_method = compose.multilib_method
        old_multilib_method = old_compose.multilib_method
        if multilib_method != old_multilib_method:
            log.debug(
                "%r: Cannot reuse %r - multilib_method not same", compose, old_compose
            )
            continue

        modular_koji_tags = (
            set(compose.modular_koji_tags.split(" "))
            if compose.modular_koji_tags
            else set()
        )
        old_modular_koji_tags = (
            set(old_compose.modular_koji_tags.split(" "))
            if old_compose.modular_koji_tags
            else set()
        )
        if modular_koji_tags != old_modular_koji_tags:
            log.debug(
                "%r: Cannot reuse %r - modular_koji_tags not same", compose, old_compose
            )
            continue

        module_defaults_url = compose.module_defaults_url
        old_module_defaults_url = old_compose.module_defaults_url
        if module_defaults_url != old_module_defaults_url:
            log.debug(
                "%r: Cannot reuse %r - module_defaults_url not same",
                compose,
                old_compose,
            )
            continue

        scratch_modules = (
            set(compose.scratch_modules.split(" "))
            if compose.scratch_modules
            else set()
        )
        old_scratch_modules = (
            set(old_compose.scratch_modules.split(" "))
            if old_compose.scratch_modules
            else set()
        )
        if scratch_modules != old_scratch_modules:
            log.debug(
                "%r: Cannot reuse %r - scratch_modules not same", compose, old_compose
            )
            continue

        scratch_build_tasks = (
            set(compose.scratch_build_tasks.split(" "))
            if compose.scratch_build_tasks
            else set()
        )
        old_scratch_build_tasks = (
            set(old_compose.scratch_build_tasks.split(" "))
            if old_compose.scratch_build_tasks
            else set()
        )
        if scratch_build_tasks != old_scratch_build_tasks:
            log.debug(
                "%r: Cannot reuse %r - scratch_build_tasks not same",
                compose,
                old_compose,
            )
            continue

        parent_pungi_compose_ids = (
            set(compose.parent_pungi_compose_ids.split(" "))
            if compose.parent_pungi_compose_ids
            else set()
        )
        old_parent_pungi_compose_ids = (
            set(old_compose.parent_pungi_compose_ids.split(" "))
            if old_compose.parent_pungi_compose_ids
            else set()
        )
        if parent_pungi_compose_ids != old_parent_pungi_compose_ids:
            log.debug(
                "%r: Cannot reuse %r - parent_pungi_compose_ids not same",
                compose,
                old_compose,
            )
            continue

        target_dir = compose.target_dir
        old_target_dir = old_compose.target_dir
        if target_dir != old_target_dir:
            log.debug("%r: Cannot reuse %r - target_dir not same", compose, old_compose)
            continue

        # In case of compose renewal, the compose.koji_event will be actually
        # lower than the "old_compose"'s one - the `compose` might have been for
        # example submitted 1 year ago, so koji_event will be one year old.
        # But the `old_compose` was submitted few days ago at max.
        # In this case, we must never reuse the newer compose for old one.
        if (
            compose.koji_event
            and old_compose.koji_event
            and compose.koji_event < old_compose.koji_event
        ):
            log.debug(
                "%r: Cannot reuse %r - koji_event of current compose "
                "is lower than koji_event of old compose.",
                compose,
                old_compose,
            )
            continue

        if compose.source_type == PungiSourceType.KOJI_TAG:
            # For KOJI_TAG compose, check that all the inherited tags by our
            # Koji tag have not changed since previous old_compose.
            koji_session = create_koji_session()
            if tag_changed(koji_session, compose.source, old_compose.koji_event):
                log.debug(
                    "%r: Cannot reuse %r - one of the tags changed "
                    "since previous compose.",
                    compose,
                    old_compose,
                )
                continue
        elif compose.koji_event != old_compose.koji_event:
            log.debug(
                "%r: Cannot reuse %r - koji_events not same, %d != %d",
                compose,
                old_compose,
                compose.koji_event,
                old_compose.koji_event,
            )
            continue

        return old_compose

    return None


def reuse_compose(compose, compose_to_reuse):
    """
    Changes the attribute of `compose` in a way it reuses
    the `compose_to_reuse`.
    """

    # Set the reuse_id
    compose.reused_id = compose_to_reuse.id
    # Set the time_to_expire to bigger value from both composes.
    compose.time_to_expire = max(
        compose.time_to_expire, compose_to_reuse.time_to_expire
    )
    # NOTE: reuse_compose is only called by generate_pungi_compose at this
    # moment. This change will be committed when compose state is transitted,
    # which will call session's commit. If this method is called from somewhere
    # else in the future, be careful to manage the commit.
    compose_to_reuse.time_to_expire = compose.time_to_expire


def _write_repo_file(compose, data=None):
    """
    Writes main repo file for a resulting compose containing the `data`.
    If `data` is not provided, the default one pointing to pungi compose
    will be generated.
    """
    if not data:
        baseurl = os.path.join(compose.result_repo_url, "$basearch", "os")
        data = """[%s]
name=ODCS repository for compose %s
baseurl=%s
type=rpm-md
skip_if_unavailable=False
gpgcheck=0
repo_gpgcheck=0
enabled=1
enabled_metadata=1
""" % (
            compose.name,
            compose.name,
            baseurl,
        )

    # Ensure the directory exists
    dirname = os.path.dirname(compose.result_repofile_path)
    odcs.server.utils.makedirs(dirname)

    with open(compose.result_repofile_path, "w") as f:
        f.write(data)


def generate_pulp_compose(compose):
    """
    Generates the "compose" of PULP type - this basically means only
    repo file pointing to data in pulp.
    """
    content_sets = compose.source.split(" ")

    pulp = Pulp(
        server_url=conf.pulp_server_url,
        username=conf.pulp_username,
        password=conf.pulp_password,
        compose=compose,
    )

    repofile = ""
    repos = pulp.get_repos_from_content_sets(
        content_sets, compose.flags & COMPOSE_FLAGS["include_unpublished_pulp_repos"]
    )
    ignore_absent_pulp_repos = compose.flags & COMPOSE_FLAGS["ignore_absent_pulp_repos"]
    if len(repos) != len(content_sets):
        found_content_sets = repos.keys()
        err = "Failed to find all the content_sets %r in Pulp, " "found only %r" % (
            content_sets,
            found_content_sets,
        )
        if ignore_absent_pulp_repos:
            log.info(err)
            # Update the source in the compose. This ensures the source matches
            # what is actually in the compose. However it makes it invisible
            # that user actually requested something else.
            compose.source = " ".join(found_content_sets)
        else:
            log.error(err)
            raise ValueError(err)

    arches = set()
    sigkeys = set()
    for name in sorted(repos.keys()):
        repo_data = repos[name]
        url = repo_data["url"]
        r = """
[%s]
name=%s
baseurl=%s
enabled=1
gpgcheck=0
""" % (
            name,
            name,
            url,
        )
        repofile += r
        arches = arches.union(repo_data["arches"])
        sigkeys = sigkeys.union(repo_data["sigkeys"])

    _write_repo_file(compose, repofile)

    compose.arches = " ".join(arches)
    compose.sigkeys = " ".join(sigkeys)
    compose.transition(COMPOSE_STATES["done"], "Compose is generated successfully")
    log.info("%r: Compose done", compose)


def generate_compose_symlink(compose):
    """
    Generates symlink(s) for compose based on its `compose.pungi_compose_id`.

    It generates following symlinks pointing to the compose:
      - $compose.target_dir/$compose.compose_type/$compose.pungi_compose_id
      - $compose.target_dir/$compose.compose_type/latest-$name-version

    If the latest-* symlink exists, it is replaced with new one pointing to
    the `composes`.
    """
    symlink_dir = os.path.join(compose.target_dir, compose.compose_type)
    odcs.server.utils.makedirs(symlink_dir)

    # Generate the non-latest symlink.
    compose_dir = os.path.relpath(compose.toplevel_dir, symlink_dir)
    symlink = os.path.join(symlink_dir, compose.pungi_compose_id)
    log.info("%r: Generating %s symlink.", compose, symlink)
    os.symlink(compose_dir, symlink)

    # Generate the latest-* symlink.
    latest_name = "latest-%s" % "-".join(compose.pungi_compose_id.split("-")[:2])
    symlink = os.path.join(symlink_dir, latest_name)
    try:
        os.unlink(symlink)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
    log.info("%r: Generating %s symlink.", compose, symlink)
    os.symlink(compose_dir, symlink)


def get_latest_symlink(compose):
    """
    Returns the latest-* symlink associated with this compose or None
    if it does not exists.
    """
    # Ignore old composes which do not have the compose_type and pungi_compose_id set.
    if not compose.compose_type or not compose.pungi_compose_id:
        return None

    symlink_dir = os.path.join(compose.target_dir, compose.compose_type)
    symlink = os.path.join(symlink_dir, compose.pungi_compose_id)
    latest_name = "latest-%s" % "-".join(compose.pungi_compose_id.split("-")[:2])
    latest_symlink = os.path.join(symlink_dir, latest_name)

    # Return Non if `latest_symlink` points to the different dir than `symlink`.
    if os.path.realpath(symlink) != os.path.realpath(latest_symlink):
        return None

    return latest_symlink


def remove_compose_symlink(compose):
    """
    Removes non-latest symlink generated by the `generate_compose_symlink`.
    """
    # Do not try to remove symlinks from old composes which do not have
    # the compose_type and pungi_compose_id set.
    if not compose.compose_type or not compose.pungi_compose_id:
        return

    symlink_dir = os.path.join(compose.target_dir, compose.compose_type)
    symlink = os.path.join(symlink_dir, compose.pungi_compose_id)
    latest_symlink = get_latest_symlink(compose)

    # Remove non-latest symlink.
    log.info("%r: Removing %s symlink.", compose, symlink)
    try:
        os.unlink(symlink)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise

    # Remove latest symlink.
    if latest_symlink:
        log.info("%r: Removing %s symlink.", compose, latest_symlink)
        try:
            os.unlink(latest_symlink)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise


def generate_pungi_compose(compose):
    """
    Generates the compose of KOJI, TAG, or REPO type using the Pungi tool.

    :param models.Compose compose: Instance of models.Compose.
    """
    koji_tag_cache = KojiTagCache(compose)

    # Resolve the general data in the compose.
    resolve_compose(compose)

    # Reformat the data from database
    packages = compose.packages
    if packages:
        packages = packages.split(" ")
    builds = compose.builds
    if builds:
        builds = builds.split(" ")

    # Check if we can reuse some existing compose instead of
    # generating new one.
    compose_to_reuse = get_reusable_compose(compose)
    if compose_to_reuse:
        log.info("%r: Reusing compose %r", compose, compose_to_reuse)
        reuse_compose(compose, compose_to_reuse)
    else:
        if compose.source_type == PungiSourceType.RAW_CONFIG:
            pungi_cfg = RawPungiConfig(compose)
        else:
            # Generate PungiConfig and run Pungi
            if compose.multilib_arches:
                multilib_arches = compose.multilib_arches.split(" ")
            else:
                multilib_arches = None
            pungi_cfg = PungiConfig(
                compose.name,
                "1",
                compose.source_type,
                compose.source,
                packages=packages,
                sigkeys=compose.sigkeys,
                results=compose.results,
                arches=compose.arches.split(" "),
                multilib_arches=multilib_arches,
                multilib_method=compose.multilib_method,
                builds=builds,
                flags=compose.flags,
                lookaside_repos=compose.lookaside_repos,
                modular_koji_tags=compose.modular_koji_tags,
                module_defaults_url=compose.module_defaults_url,
                scratch_modules=compose.scratch_modules,
                scratch_build_tasks=compose.scratch_build_tasks,
            )
            if compose.flags & COMPOSE_FLAGS["no_deps"]:
                pungi_cfg.gather_method = "nodeps"
            if compose.flags & COMPOSE_FLAGS["no_inheritance"]:
                pungi_cfg.pkgset_koji_inherit = False
            if compose.flags & COMPOSE_FLAGS["no_reuse"]:
                pungi_cfg.pkgset_allow_reuse = False

        koji_event = None
        if compose.source_type == PungiSourceType.KOJI_TAG:
            koji_event = compose.koji_event

        old_compose = None
        if koji_tag_cache.is_cached(compose) and pungi_cfg.pkgset_allow_reuse:
            koji_tag_cache.reuse_cached(compose)
            old_compose = koji_tag_cache.cache_dir

        # For Raw config composes, ODCS generates symlinks named according
        # to Pungi COMPOSE_ID. We can use directory with these symlinks to
        # find out previous old compose.
        if compose.source_type == PungiSourceType.RAW_CONFIG:
            old_compose = os.path.join(compose.target_dir, compose.compose_type)

        pungi = Pungi(compose.id, pungi_cfg, koji_event, old_compose)
        pungi.run(compose)

        # Write the Compose.result_repofile_path. This makes sense only for
        # non raw_config composes, because we have no idea what variants
        # do exist for raw_config composes.
        if compose.source_type != PungiSourceType.RAW_CONFIG:
            _write_repo_file(compose)

    # Raises an exception if invalid
    validate_pungi_compose(compose)

    # Generate symlinks pointing to latest build of raw_config compose.
    if compose.source_type == PungiSourceType.RAW_CONFIG:
        generate_compose_symlink(compose)
        if compose.compose_type in ["production", "development"]:
            pungi_logs = PungiLogs(compose)
            config_dump = pungi_logs.get_config_dump()
            if not config_dump:
                msg = "%r: Cannot load Pungi config dump." % compose
                log.error(msg)
                raise RuntimeError(msg)
            compose.pungi_config_dump = config_dump

    # If there is no exception generated by the pungi.run() and if
    # validation didn't fail, then we know the compose has been
    # successfully generated.
    compose.transition(COMPOSE_STATES["done"], "Compose is generated successfully")
    log.info("%r: Compose done", compose)

    koji_tag_cache.update_cache(compose)


def validate_pungi_compose(compose):
    """
    Validate the compose is generated by pungi as expected.
    """
    # the requested packages should be present in the generated compose
    if compose.packages:
        packages = compose.packages.split()
        pungi_compose = productmd.compose.Compose(compose.toplevel_dir)
        rm = pungi_compose.rpms.rpms
        rpm_nevras = []
        for variant in rm:
            for arch in rm[variant]:
                for srpm_nevra, data in six.iteritems(rm[variant][arch]):
                    for rpm_nevra, data in six.iteritems(rm[variant][arch][srpm_nevra]):
                        if data["category"] == "source":
                            continue
                        rpm_nevras.append(rpm_nevra)
        rpms = set([productmd.common.parse_nvra(n)["name"] for n in rpm_nevras])
        not_found = []
        for pkg in packages:
            if pkg not in rpms:
                not_found.append(pkg)
        if not_found:
            msg = (
                "The following requested packages are not present in the generated compose: %s."
                % " ".join(not_found)
            )
            log.error(msg)
            raise RuntimeError(msg)


def generate_compose(compose_id, lost_compose=False):
    """
    Generates the compose defined by its `compose_id`. It is run by
    ThreadPoolExecutor from the ComposerThread.
    """
    compose = None
    with app.app_context():
        try:
            # Get the compose from database.
            compose = Compose.query.filter(Compose.id == compose_id).one()
            log.info("%r: Starting compose generation", compose)

            if compose.source_type == PungiSourceType.PULP:
                # Pulp compose is special compose not generated by Pungi.
                # The ODCS in this case just creates .repo file which points
                # to composes generated by pulp/pub.
                generate_pulp_compose(compose)
            else:
                generate_pungi_compose(compose)
        except Exception as e:
            # Something went wrong, log the exception and update the compose
            # state in database.
            if compose:
                log.exception("%r: Error while generating compose", compose)
            else:
                log.exception("Error while generating compose %d", compose_id)

            state_reason = "Error while generating compose: {}".format(str(e))

            try:
                pungi_logs = PungiLogs(compose)
                state_reason += "\n{}".format(pungi_logs.get_error_string())
            except Exception:
                log.exception("Exception raised when getting Pungi logs.")

            # Be nice to end user and replace paths to logs or other files with URL
            # accessible to the user.
            if compose.on_default_target_dir:
                state_reason = state_reason.replace(
                    conf.target_dir, conf.target_dir_url
                )
            compose.transition(COMPOSE_STATES["failed"], state_reason)

        compose = Compose.query.filter(Compose.id == compose_id).one()

        koji_tag_cache = KojiTagCache(compose)
        koji_tag_cache.cleanup_reused(compose)

        # Commit the session to ensure that database transaction is closed and
        # does not remain in Idle state acquiring the table lock.
        db.session.commit()


class ComposerThread(BackendThread):
    """
    Thread used to query the database for composes in "wait" state and
    generating the composes using Pungi.
    """

    def __init__(self):
        """
        Creates new ComposerThread instance.
        """
        super(ComposerThread, self).__init__(1)
        # ThreadPool to execute composes which calls "pungi" process.
        self.executor = ThreadPoolExecutor(conf.num_concurrent_pungi)
        # Threadpool to execute Pulp composes. These composes take just
        # few seconds and do just single HTTP call to Pulp, so there is
        # no need to put them into the same queue as Pungi based composes.
        self.pulp_executor = ThreadPoolExecutor(2)

        # List of composes which are being currently generated by this ODCS
        # instance.
        self.currently_generating = []

    def generate_new_compose(self, compose):
        """
        Adds the compose to queue of composes to generate, so
        the ThreadPoolExecutor can start working on it.
        """
        compose.transition(COMPOSE_STATES["generating"], "Compose thread started")

        self.currently_generating.append(compose.id)
        if compose.source_type == PungiSourceType.PULP:
            self.pulp_executor.submit(generate_compose, compose.id)
        else:
            self.executor.submit(generate_compose, compose.id)

    def generate_new_composes(self):
        """
        Gets all the composes in "wait" state. Generates them using Pungi
        by calling `generate_compose(...)` in ThreadPoolExecutor.
        """
        composes = Compose.query.filter(Compose.state == COMPOSE_STATES["wait"]).all()

        for compose in composes:
            log.info("%r: Going to start compose generation.", compose)
            self.generate_new_compose(compose)

    def fail_lost_generating_composes(self):
        """
        Fails the composes in `generating` state in case they are in this
        state for longer than `2 * conf.pungi_timeout`. Because composes
        can be generating only for `conf.pungi_timeout` seconds, this is enough
        time to generate any compose.
        """
        max_generating_time = 2 * conf.pungi_timeout
        now = datetime.utcnow()
        too_old_datetime = now - timedelta(seconds=max_generating_time)

        # Get composes which are in 'generating' state for too long.
        composes = (
            Compose.query.filter(
                Compose.state == COMPOSE_STATES["generating"],
                Compose.time_started < too_old_datetime,
            )
            .order_by(Compose.id)
            .all()
        )

        for compose in composes:
            compose.transition(
                COMPOSE_STATES["failed"],
                "Compose stuck in 'generating' state for longer than %d "
                "seconds." % max_generating_time,
            )

    def generate_lost_composes(self):
        """
        Gets all the composes in "generating" state and continues with
        the generation process.

        This method is here to handle situation where the ODCS is restarted
        in the middle of compose generation.
        """
        composes = Compose.query.filter(
            Compose.state == COMPOSE_STATES["generating"]
        ).all()

        for compose in composes:
            if compose.id in self.currently_generating:
                # We already have a thread working on this compose.
                continue

            log.info("%r: Going to regenerate lost compose.", compose)
            self.generate_new_compose(compose)

    def refresh_currently_generating(self):
        """
        Checks the status of all composes in self.currently_generating
        and removes those which have been already done from this list.
        """

        new_currently_generating_list = []
        for compose_id in self.currently_generating:
            compose = Compose.query.filter(Compose.id == compose_id).one()
            if compose.state != COMPOSE_STATES["generating"]:
                continue

            new_currently_generating_list.append(compose_id)
        self.currently_generating = new_currently_generating_list

    def do_work(self):
        """
        Gets all the composes in "wait" state. Generates them using Pungi
        by calling `generate_compose(...)` in ThreadPoolExecutor.
        """
        self.generate_lost_composes()
        self.generate_new_composes()
        self.refresh_currently_generating()



"""Create user model

Revision ID: 0d4d8e1cfe29
Revises: 566733ac3811
Create Date: 2017-07-06 22:37:38.576275

"""

# revision identifiers, used by Alembic.
revision = "0d4d8e1cfe29"
down_revision = "566733ac3811"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "users",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("username", sa.String(length=200), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("username"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("users")
    # ### end Alembic commands ###



"""Add sigkeys field.

Revision ID: b75ad2afc207
Revises: c370b90de998
Create Date: 2017-10-03 13:41:42.303962

"""

# revision identifiers, used by Alembic.
revision = "b75ad2afc207"
down_revision = "c370b90de998"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("composes", sa.Column("sigkeys", sa.String(), nullable=True))


def downgrade():
    op.drop_column("composes", "sigkeys")



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

from functools import wraps
import json
import responses
from six.moves.urllib.parse import urlparse, parse_qs

from odcs.server import conf

import gi

gi.require_version("Modulemd", "2.0")
from gi.repository import Modulemd  # noqa: E402


def dump_mmd(mmd):
    mod_index = Modulemd.ModuleIndex.new()
    mod_index.add_module_stream(mmd)
    return mod_index.dump_to_string()


def make_module(name, stream, version, requires={}, mdversion=1, context=None, state=5):
    if mdversion == 1:
        mmd = Modulemd.ModuleStreamV1.new(name, stream)
    else:
        mmd = Modulemd.ModuleStreamV2.new(name, stream)
    mmd.set_version(version)
    mmd.set_context(context or "00000000")
    mmd.set_summary("foo")
    mmd.set_description("foo")
    mmd.add_module_license("GPL")

    if mdversion == 1:
        for req_name, req_stream in requires.items():
            mmd.add_runtime_requirement(req_name, req_stream)
    else:
        deps = Modulemd.Dependencies()
        for req_name, req_stream in requires.items():
            deps.add_runtime_stream(req_name, req_stream)
        mmd.add_dependencies(deps)

    return {
        "name": name,
        "stream": stream,
        "version": str(version),
        "context": context or "00000000",
        "modulemd": dump_mmd(mmd),
        "state": state,
    }


TEST_MBS_MODULES_MMDv1 = [
    # test_backend.py
    make_module("moduleA", "f26", 20170809000000, {"moduleB": "f26"}),
    make_module("moduleA", "f26", 20170805000000, {"moduleB": "f26"}),
    make_module("moduleB", "f26", 20170808000000, {"moduleC": "f26", "moduleD": "f26"}),
    make_module("moduleB", "f27", 2017081000000, {"moduleC": "f27"}),
    make_module("moduleC", "f26", 20170807000000, {"moduleD": "f26"}),
    make_module("moduleD", "f26", 20170806000000),
    # test_composerthread.py
    make_module("testmodule", "master", 20170515074418),
    make_module("testmodule", "master", 20170515074419),
]


TEST_MBS_MODULES_MMDv2 = [
    # test_backend.py
    make_module("moduleA", "f26", 20170809000000, {"moduleB": "f26"}, 2),
    make_module("moduleA", "f26", 20170805000000, {"moduleB": "f26"}, 2),
    make_module(
        "moduleB", "f26", 20170808000000, {"moduleC": "f26", "moduleD": "f26"}, 2
    ),
    make_module("moduleB", "f27", 2017081000000, {"moduleC": "f27"}, 2),
    make_module("moduleC", "f26", 20170807000000, {"moduleD": "f26"}, 2),
    make_module("moduleD", "f26", 20170806000000, {}, 2),
    # module builds in "done" state.
    make_module("testmodule", "master", 20180515074419, {}, 2, state=3),
    # test_composerthread.py
    make_module("testmodule", "master", 20170515074418, {}, 2),
    make_module("testmodule", "master", 20170515074419, {}, 2),
    # multiple contexts
    make_module("parent", "master", 1, {}, 2, context="a"),
    make_module("parent", "master", 1, {}, 2, context="b"),
    make_module("testcontexts", "master", 1, {"parent": "master"}, 2, context="a"),
    make_module("testcontexts", "master", 1, {"parent": "master"}, 2, context="b"),
]


def mock_mbs(mdversion=2):
    """
    Decorator that sets up a test environment so that calls to the MBS to look
    up modules are redirected to return results from the TEST_MODULES array
    above.
    """

    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            def handle_module_builds(request):
                query = parse_qs(urlparse(request.url).query)
                states = [int(s) for s in query["state"]]
                if "nsvc" in query:
                    nsvc = query["nsvc"][0]
                    nsvc_parts = nsvc.split(":")
                    nsvc_keys = ["name", "stream", "version", "context"]
                    nsvc_dict = {}
                    for key, part in zip(nsvc_keys, nsvc_parts):
                        nsvc_dict[key] = part
                else:
                    # Empty keys and dict in case nsvc is not specitified.
                    # This means no filtering based on the nsvc will be done.
                    nsvc_keys = []
                    nsvc_dict = {}

                if mdversion == 1:
                    modules = TEST_MBS_MODULES_MMDv1
                else:
                    modules = TEST_MBS_MODULES_MMDv2

                body = {"items": [], "meta": {"total": 0}}
                for module in modules:
                    skip = False
                    for key in nsvc_keys:
                        if key in nsvc_dict and nsvc_dict[key] != module[key]:
                            skip = True
                            break
                    if module["state"] not in states:
                        skip = True
                    if skip:
                        continue
                    body["items"].append(module)

                body["meta"]["total"] = len(body["items"])
                return (200, {}, json.dumps(body))

            responses.add_callback(
                responses.GET,
                conf.mbs_url + "/1/module-builds/",
                content_type="application/json",
                callback=handle_module_builds,
            )

            return f(*args, **kwargs)

        return responses.activate(wrapped)

    return wrapper



"""empty message

Revision ID: 566733ac3811
Revises: None
Create Date: 2017-06-08 15:50:27.348945

"""

# revision identifiers, used by Alembic.
revision = "566733ac3811"
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "composes",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("owner", sa.String(), nullable=False),
        sa.Column("source_type", sa.Integer(), nullable=False),
        sa.Column("source", sa.String(), nullable=False),
        sa.Column("koji_event", sa.Integer(), nullable=True),
        sa.Column("state", sa.Integer(), nullable=False),
        sa.Column("results", sa.Integer(), nullable=False),
        sa.Column("packages", sa.String(), nullable=True),
        sa.Column("flags", sa.Integer(), nullable=True),
        sa.Column("reused_id", sa.Integer(), nullable=True),
        sa.Column("time_to_expire", sa.DateTime(), nullable=False),
        sa.Column("time_submitted", sa.DateTime(), nullable=False),
        sa.Column("time_done", sa.DateTime(), nullable=True),
        sa.Column("time_removed", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("composes")
    ### end Alembic commands ###



"""Add index on Compose.reused_id

Revision ID: f24a36cc8a16
Revises: 3b92820da295
Create Date: 2017-09-21 11:42:00.622541

"""

# revision identifiers, used by Alembic.
revision = "f24a36cc8a16"
down_revision = "3b92820da295"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        op.f("ix_composes_reused_id"), "composes", ["reused_id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_composes_reused_id"), table_name="composes")
    # ### end Alembic commands ###
